from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional
from uuid import UUID, uuid4

from pfmsoft import aiohttp_queue
from pfmsoft.aiohttp_queue import callbacks as AQ_callbacks


@dataclass
class TestAction:
    action: aiohttp_queue.AiohttpAction
    context: Dict = field(default_factory=dict)


def get_with_response_text(params: Dict[str, str]) -> TestAction:
    url_template = "https://httpbin.org/get"
    url_params = None
    callbacks = aiohttp_queue.ActionCallbacks(
        success=[AQ_callbacks.ResponseContentToText()]
    )

    test_action = build_get_test_action(
        "get_with_response_text",
        url_template=url_template,
        url_parameters=url_params,
        params=params,
        callbacks=callbacks,
    )
    return test_action


def get_with_response_json(params: Dict[str, str]) -> TestAction:
    url_template = "https://httpbin.org/get"
    url_params = None
    callbacks = aiohttp_queue.ActionCallbacks(
        success=[AQ_callbacks.ResponseContentToJson()]
    )
    test_action = build_get_test_action(
        "get_with_response_json",
        url_template=url_template,
        url_parameters=url_params,
        params=params,
        callbacks=callbacks,
    )
    return test_action


def get_with_404() -> TestAction:
    test_action = None
    return test_action


def get_with_501() -> TestAction:
    test_action = None
    return test_action


def get_list_of_dicts_result(url_params: Dict, params: Dict) -> TestAction:
    url_template = "https://esi.evetech.net/latest/markets/${region_id}/history"
    callbacks = aiohttp_queue.ActionCallbacks(
        success=[AQ_callbacks.ResponseContentToJson()]
    )
    test_action = build_get_test_action(
        "get_list_of_dicts_result",
        url_template=url_template,
        url_parameters=url_params,
        params=params,
        callbacks=callbacks,
    )
    return test_action


def build_get_test_action(name, url_template, url_parameters, params, callbacks):
    request_qwargs = {"params": params}
    action = aiohttp_queue.AiohttpAction(
        "get",
        url_template=url_template,
        url_parameters=url_parameters,
        request_kwargs=request_qwargs,
        name=name,
        id_=uuid4(),
        callbacks=callbacks,
    )
    test_action = TestAction(
        action,
        {
            "url_template": url_template,
            "url_parameters": url_parameters,
            "params": params,
        },
    )
    return test_action


def save_list_of_dicts_to_csv_file(
    url_params: Dict[str, str],
    params: Dict[str, str],
    file_path: Path,
    template_params: Optional[Dict] = None,
    file_ending: str = ".csv",
):
    test_action = get_list_of_dicts_result(url_params, params)
    test_action.action.callbacks.success.append(
        AQ_callbacks.SaveListOfDictResultToCSVFile(
            file_path=file_path,
            template_params=template_params,
            file_ending=file_ending,
        )
    )
    return test_action


def save_txt_to_file(
    params: Dict[str, str],
    file_path: Path,
    template_params: Optional[Dict] = None,
    file_ending: str = "",
) -> TestAction:

    test_action = get_with_response_text(params)
    test_action.action.callbacks.success.append(
        AQ_callbacks.SaveResultToFile(
            file_path=file_path,
            template_params=template_params,
            file_ending=file_ending,
        )
    )
    return test_action


def save_json_to_file(
    params: Dict[str, str],
    file_path: Path,
    template_params: Optional[Dict] = None,
    file_ending: str = ".json",
) -> TestAction:
    test_action = get_with_response_json(params)
    test_action.action.callbacks.success.append(
        AQ_callbacks.SaveJsonResultToFile(
            file_path=file_path,
            template_params=template_params,
            file_ending=file_ending,
        )
    )
    return test_action


# def action_with_data() -> TestAction:
#     test_action = None
#     return test_action
